#ifndef DOUBLELL_H_
#define DOUBLELL_H_

//vector<int> myVector;
//DLink<int> myIntLink;
//DLink<std::string> myStringLink;

template <class T>
struct DLink
{
  T data;
  DLink<T>* prev;
  DLink<T>* next;
};

template <class T> 
class DoubleLL
{
 private:
  DLink<T>* head;
  DLink<T>* tail;

 public:
  DoubleLL(); // creates empty list
  ~DoubleLL(); // deallocates any leftover memory

  DLink<T>* getHead() const;
  DLink<T>* getTail() const;
  
  // inserts a new string at the spot _before_ 'where'.
  // if 'where' is NULL, inserts at the end
  void insert(DLink<T>* where, const T& what);
  
  // removes the link pointed to by 'where'
  // returns the string in that link
  T remove(DLink<T>* where);

  // returns the size of the list
  int size() const;

  // gets the nth string in the list, counting from 0
  T nth(int n) const;
};


template <class T>
DoubleLL<T>::DoubleLL()
{
  head = tail = NULL;
}

template <class T>
DoubleLL<T>::~DoubleLL()
{
  while(head)
  {
    remove(head);
  }
}

template <class T>
DLink<T>* DoubleLL<T>::getHead() const
{
  return head;
}

template <class T>
DLink<T>* DoubleLL<T>::getTail() const
{
  return tail;
}

template <class T>
void DoubleLL<T>::insert(DLink<T>* where, const T& what)
{
  if(where == NULL) // adding to end
  {
    if(head == NULL) // empty list
    {
      head = tail = new DLink<T>;
      head->data = what;
      head->prev = NULL;
      head->next = NULL;
    }
    else
    {
      tail->next = new DLink<T>;
      tail->next->prev = tail;
      tail->next->next = NULL;
      tail->next->data = what;
      tail = tail->next;
    }
  }
  else
  {
    if(where == head)
    {
      head->prev = new DLink<T>;
      head->prev->prev = NULL;
      head->prev->next = head;
      head->prev->data = what;
      head = head->prev;
    }
    else
    {
      DLink<T>* link = new DLink<T>;
      link->prev = where->prev;
      link->next = where;
      link->prev->next = link;
      where->prev = link;
      link->data = what;
    }
  }
}

template <class T>
T DoubleLL<T>::remove(DLink<T>* where)
{
  T retVal = where->data;

  if(head == tail) // last element
  {
    head = tail = NULL;
  }
  else if(where == head)
  {
    head = head->next;
    head->prev = NULL;
  }
  else if(where == tail)
  {
    tail = tail->prev;
    tail->next = NULL;
  }
  else
  {
    where->prev->next = where->next;
    where->next->prev = where->prev;
  }

  delete where;
  return retVal;
}

template <class T>
int DoubleLL<T>::size() const
{
  int count = 0;
  for(DLink<T>* cur = head; cur; cur = cur->next)
  {
    count++;
  }

  return count;
}

template <class T>
T DoubleLL<T>::nth(int n) const
{
  DLink<T>* cur = head;
  for( ; n; n--)
  {
    cur = cur->next;
  }

  return cur->data;
}

#endif
